// @ts-nocheck - Skip TypeScript checking until UniversalNumber implementation is completed
/**
 * PrimeMath module for the UOR Math-JS library
 * Provides a namespace with advanced arithmetic and number theory functions
 * Based on the Prime Framework for universal number representation
 * @module PrimeMath
 */
const { PrimeMathError, toBigInt, gcd: euclideanGcd, fastExp, isPrime: isSimplePrime, nextPrime: findNextPrime, getNthPrime, isMersennePrime, moebiusFunction, quadraticResidue } = require('./Utils');
const { factorizeOptimal, millerRabinTest, fromPrimeFactors, factorMapToArray, 
// eslint-disable-next-line no-unused-vars
pollardRho } = require('./Factorization');
// Import for type checking, will be properly integrated when UniversalNumber is implemented
/**
 * @typedef {Object} UniversalNumber
 * @property {function} add - Add method
 * @property {function} subtract - Subtract method
 * @property {function} multiply - Multiply method
 * @property {function} divide - Divide method
 * @property {function} pow - Power method
 * @property {function} gcd - GCD method
 * @property {function} lcm - LCM method
 * @property {function} isIntrinsicPrime - Check if number is prime
 * @property {function} toBigInt - Convert to BigInt
 * @property {function} getFactorization - Get prime factorization
 */
/** @type {Object|null} */
let UniversalNumber = null;
try {
    // @ts-ignore - Module may not exist yet
    UniversalNumber = require('./UniversalNumber');
}
catch (e) {
    // UniversalNumber module not yet available
    UniversalNumber = null;
}
/**
 * Checks if a value is potentially a UniversalNumber
 * @private
 * @param {*} value - The value to check
 * @returns {boolean} True if the value is a UniversalNumber
 */
function isUniversalNumber(value) {
    return UniversalNumber !== null && value instanceof UniversalNumber;
}
/**
 * Extracts the prime factorization from a value if possible
 * @private
 * @param {*} value - The value to extract factorization from
 * @returns {Map<BigInt,BigInt>|null} The prime factorization or null if not available
 */
function extractFactorization(value) {
    if (isUniversalNumber(value)) {
        return value.getFactorization();
    }
    return null;
}
/**
 * Computes the coherence inner product between two factorizations
 * This measures the alignment between two numbers in the Prime Framework's reference fiber algebra
 *
 * @private
 * @param {Map<BigInt, BigInt>} factorizationA - Prime factorization of first number
 * @param {Map<BigInt, BigInt>} factorizationB - Prime factorization of second number
 * @returns {BigInt} The coherence inner product value
 */
function computeCoherenceInnerProduct(factorizationA, factorizationB) {
    let innerProduct = 0n;
    // Get all primes from both factorizations
    const allPrimes = new Set([...factorizationA.keys(), ...factorizationB.keys()]);
    // Sum the product of corresponding exponents for each prime
    for (const prime of allPrimes) {
        const exponentA = factorizationA.get(prime) || 0n;
        const exponentB = factorizationB.get(prime) || 0n;
        innerProduct += exponentA * exponentB * prime;
    }
    return innerProduct;
}
/**
 * Computes the coherence norm of a factorization
 * Measures the "magnitude" of a number in the Prime Framework's reference fiber algebra
 *
 * @private
 * @param {Map<BigInt, BigInt>} factorization - Prime factorization
 * @returns {BigInt} The coherence norm value
 */
function computeCoherenceNorm(factorization) {
    return computeCoherenceInnerProduct(factorization, factorization);
}
/**
 * The PrimeMath namespace with static arithmetic and number theory functions
 * Aligns with the Prime Framework by leveraging prime factorization for operations
 * @namespace PrimeMath
 */
const PrimeMath = {
    /**
     * Add two numbers
     * Uses regular addition for numeric types and coordinates-based addition for UniversalNumbers
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {BigInt} Sum of a and b
     */
    add(a, b) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a)) {
            return a.add(b);
        }
        if (isUniversalNumber(b)) {
            return b.add(a);
        }
        // Regular numeric addition
        const bigA = toBigInt(a);
        const bigB = toBigInt(b);
        return bigA + bigB;
    },
    /**
     * Subtract one number from another
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number (minuend)
     * @param {number|string|BigInt|UniversalNumber} b - Second number (subtrahend)
     * @returns {BigInt} Difference of a - b
     */
    subtract(a, b) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a)) {
            return a.subtract(b);
        }
        if (isUniversalNumber(b)) {
            const bigA = toBigInt(a);
            // Create a UniversalNumber if available, otherwise just return the calculation
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(bigA).subtract(b);
            }
            return bigA;
        }
        // Regular numeric subtraction
        const bigA = toBigInt(a);
        const bigB = toBigInt(b);
        return bigA - bigB;
    },
    /**
     * Multiply two numbers
     * For factorized numbers, multiplication is performed by combining their prime exponent maps
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {BigInt|UniversalNumber} Product of a and b
     */
    multiply(a, b) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a)) {
            return a.multiply(b);
        }
        if (isUniversalNumber(b)) {
            return b.multiply(a);
        }
        // Extract factorizations if available
        const factorizationA = extractFactorization(a);
        const factorizationB = extractFactorization(b);
        // If both have factorizations, combine them for efficient multiplication
        if (factorizationA && factorizationB) {
            const resultFactorization = new Map(factorizationA);
            // Combine by adding exponents for each prime factor
            for (const [prime, exponent] of factorizationB.entries()) {
                const currentExponent = resultFactorization.get(prime) || 0n;
                resultFactorization.set(prime, currentExponent + exponent);
            }
            return fromPrimeFactors(resultFactorization);
        }
        // Regular numeric multiplication
        const bigA = toBigInt(a);
        const bigB = toBigInt(b);
        return bigA * bigB;
    },
    /**
     * Perform exact division of one number by another
     * Only succeeds if the division is exact (no remainder)
     * For factorized numbers, division is performed by subtracting prime exponents
     *
     * @param {number|string|BigInt|UniversalNumber} a - Dividend
     * @param {number|string|BigInt|UniversalNumber} b - Divisor
     * @returns {BigInt|UniversalNumber} Result of a / b
     * @throws {PrimeMathError} If division is not exact or divisor is zero
     */
    divide(a, b) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a)) {
            return a.divide(b);
        }
        if (isUniversalNumber(b)) {
            const bigA = toBigInt(a);
            // Create a UniversalNumber if available, otherwise just return the calculation
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(bigA).divide(b);
            }
            return bigA;
        }
        // Extract factorizations if available
        const factorizationA = extractFactorization(a);
        const factorizationB = extractFactorization(b);
        const bigA = toBigInt(a);
        const bigB = toBigInt(b);
        if (bigB === 0n) {
            throw new PrimeMathError('Division by zero is not allowed');
        }
        // If both have factorizations, perform division by subtracting exponents
        if (factorizationA && factorizationB) {
            const resultFactorization = new Map(factorizationA);
            let isExact = true;
            // Subtract exponents for each prime factor
            for (const [prime, exponent] of factorizationB.entries()) {
                const currentExponent = resultFactorization.get(prime) || 0n;
                if (currentExponent < exponent) {
                    isExact = false;
                    break;
                }
                resultFactorization.set(prime, currentExponent - exponent);
                // Remove factors with zero exponent
                if (resultFactorization.get(prime) === 0n) {
                    resultFactorization.delete(prime);
                }
            }
            if (!isExact) {
                throw new PrimeMathError(`${bigA} is not divisible by ${bigB}`);
            }
            return fromPrimeFactors(resultFactorization);
        }
        // Regular division with exactness check
        if (bigA % bigB !== 0n) {
            throw new PrimeMathError(`${bigA} is not divisible by ${bigB}`);
        }
        return bigA / bigB;
    },
    /**
     * Raise a number to a power
     * For factorized numbers, exponentiation is performed by multiplying prime exponents
     *
     * @param {number|string|BigInt|UniversalNumber} base - The base
     * @param {number|string|BigInt} exponent - The exponent (must be non-negative)
     * @returns {BigInt|UniversalNumber} base^exponent
     * @throws {PrimeMathError} If exponent is negative
     */
    pow(base, exponent) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(base)) {
            return base.pow(exponent);
        }
        const bigExponent = toBigInt(exponent);
        if (bigExponent < 0n) {
            throw new PrimeMathError('Exponent must be non-negative');
        }
        if (bigExponent === 0n) {
            return 1n;
        }
        // Extract factorization if available
        const factorization = extractFactorization(base);
        if (factorization && bigExponent > 0n) {
            const resultFactorization = new Map();
            // Multiply each prime's exponent by the power
            for (const [prime, exp] of factorization.entries()) {
                resultFactorization.set(prime, exp * bigExponent);
            }
            return fromPrimeFactors(resultFactorization);
        }
        // Regular exponentiation
        const bigBase = toBigInt(base);
        return fastExp(bigBase, bigExponent);
    },
    /**
     * Find the greatest common divisor (GCD) of two numbers
     * For factorized numbers, GCD is computed by taking the minimum of each prime's exponents
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {BigInt|UniversalNumber} The GCD of a and b
     */
    gcd(a, b) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a)) {
            return a.gcd(b);
        }
        if (isUniversalNumber(b)) {
            return b.gcd(a);
        }
        // Extract factorizations if available
        const factorizationA = extractFactorization(a);
        const factorizationB = extractFactorization(b);
        const bigA = toBigInt(a);
        const bigB = toBigInt(b);
        // If both numbers are zero, the GCD is defined as zero
        if (bigA === 0n && bigB === 0n) {
            return 0n;
        }
        // If one number is zero, the GCD is the absolute value of the other
        if (bigA === 0n) {
            return bigB < 0n ? -bigB : bigB;
        }
        if (bigB === 0n) {
            return bigA < 0n ? -bigA : bigA;
        }
        // If both have factorizations, compute GCD using prime exponents
        if (factorizationA && factorizationB) {
            const resultFactorization = new Map();
            // Find common primes and take the minimum of the exponents
            const allPrimes = new Set([...factorizationA.keys(), ...factorizationB.keys()]);
            for (const prime of allPrimes) {
                const exponentA = factorizationA.get(prime) || 0n;
                const exponentB = factorizationB.get(prime) || 0n;
                const minExponent = exponentA < exponentB ? exponentA : exponentB;
                if (minExponent > 0n) {
                    resultFactorization.set(prime, minExponent);
                }
            }
            return fromPrimeFactors(resultFactorization);
        }
        // Fallback to Euclidean algorithm
        return euclideanGcd(bigA, bigB);
    },
    /**
     * Find the least common multiple (LCM) of two numbers
     * For factorized numbers, LCM is computed by taking the maximum of each prime's exponents
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {BigInt|UniversalNumber} The LCM of a and b
     */
    lcm(a, b) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a)) {
            return a.lcm(b);
        }
        if (isUniversalNumber(b)) {
            return b.lcm(a);
        }
        // Extract factorizations if available
        const factorizationA = extractFactorization(a);
        const factorizationB = extractFactorization(b);
        const bigA = toBigInt(a);
        const bigB = toBigInt(b);
        // If either number is zero, the LCM is defined as zero
        if (bigA === 0n || bigB === 0n) {
            return 0n;
        }
        // Take absolute values for calculation
        const absA = bigA < 0n ? -bigA : bigA;
        const absB = bigB < 0n ? -bigB : bigB;
        // If both have factorizations, compute LCM using prime exponents
        if (factorizationA && factorizationB) {
            const resultFactorization = new Map();
            // Find all primes and take the maximum of the exponents
            const allPrimes = new Set([...factorizationA.keys(), ...factorizationB.keys()]);
            for (const prime of allPrimes) {
                const exponentA = factorizationA.get(prime) || 0n;
                const exponentB = factorizationB.get(prime) || 0n;
                const maxExponent = exponentA > exponentB ? exponentA : exponentB;
                if (maxExponent > 0n) {
                    resultFactorization.set(prime, maxExponent);
                }
            }
            return fromPrimeFactors(resultFactorization);
        }
        // Fallback to using the GCD
        // The result should always be positive
        return (absA / euclideanGcd(absA, absB)) * absB;
    },
    /**
     * Check if a number is prime
     * Uses the Miller-Rabin primality test for larger numbers
     * For UniversalNumber, leverages the intrinsic factorization
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to check
     * @param {Object} [options] - Options for primality testing
     * @param {boolean} [options.advanced=true] - Use advanced primality testing for large numbers
     * @returns {boolean} True if n is prime, false otherwise
     */
    isPrime(n, options = {}) {
        // Handle UniversalNumber directly if available
        if (isUniversalNumber(n)) {
            return n.isIntrinsicPrime();
        }
        const num = toBigInt(n);
        const { advanced = true } = options;
        if (num <= 1n) {
            return false;
        }
        // Check for existing factorization
        const factorization = extractFactorization(n);
        if (factorization) {
            // A number is prime if and only if it has exactly one prime factor with exponent 1
            return factorization.size === 1 && [...factorization.values()][0] === 1n;
        }
        // Use simple primality test for small numbers
        if (num < 1000000n) {
            return isSimplePrime(num);
        }
        // Use Miller-Rabin test for larger numbers if advanced option is enabled
        if (advanced) {
            return millerRabinTest(num);
        }
        // Fall back to simple primality test if advanced is false
        return isSimplePrime(num);
    },
    /**
     * Find the next prime number after a given number
     *
     * @param {number|string|BigInt|UniversalNumber} n - The starting number
     * @returns {BigInt|UniversalNumber} The next prime number after n
     */
    nextPrime(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const nextPrimeValue = findNextPrime(n.toBigInt());
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(nextPrimeValue);
            }
            return nextPrimeValue;
        }
        const num = toBigInt(n);
        const nextPrimeValue = findNextPrime(num);
        return nextPrimeValue;
    },
    /**
     * Return the nth prime number
     * Provides direct access to number theory sequences in the Prime Framework
     *
     * @param {number|string|BigInt} n - The index (1-based) of the prime to retrieve
     * @returns {BigInt|UniversalNumber} The nth prime number
     * @throws {PrimeMathError} If n is not a positive integer
     */
    nthPrime(n) {
        const index = toBigInt(n);
        if (index <= 0n) {
            throw new PrimeMathError('Index must be a positive integer');
        }
        const result = getNthPrime(index);
        if (UniversalNumber) {
            // @ts-ignore - UniversalNumber constructor is properly implemented
            return new UniversalNumber(result);
        }
        return result;
    },
    /**
     * Compute the primorial of n (the product of all primes ≤ n)
     *
     * @param {number|string|BigInt|UniversalNumber} n - Upper limit
     * @returns {BigInt|UniversalNumber} The primorial of n
     */
    primorial(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const nValue = n.toBigInt();
            const primorialValue = this.primorial(nValue);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(primorialValue);
            }
            return primorialValue;
        }
        const num = toBigInt(n);
        if (num < 2n) {
            return 1n;
        }
        // Build the primorial as a factorization
        const factorization = new Map();
        let prime = 2n;
        while (prime <= num) {
            factorization.set(prime, 1n);
            prime = findNextPrime(prime);
        }
        return fromPrimeFactors(factorization);
    },
    /**
     * Calculate the modular inverse of a number (a^-1 mod m)
     *
     * @param {number|string|BigInt|UniversalNumber} a - The number to find the inverse for
     * @param {number|string|BigInt|UniversalNumber} m - The modulus
     * @returns {BigInt|UniversalNumber|null} The modular inverse, or null if it doesn't exist
     * @throws {PrimeMathError} If the modulus is not positive
     */
    modInverse(a, m) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a) || isUniversalNumber(m)) {
            const aValue = isUniversalNumber(a) ? a.toBigInt() : toBigInt(a);
            const mValue = isUniversalNumber(m) ? m.toBigInt() : toBigInt(m);
            const result = this.modInverse(aValue, mValue);
            if (result === null) {
                return null;
            }
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        let bigA = toBigInt(a);
        const bigM = toBigInt(m);
        if (bigM <= 0n) {
            throw new PrimeMathError('Modulus must be positive');
        }
        // Ensure a is positive and within the range of the modulus
        bigA = ((bigA % bigM) + bigM) % bigM;
        if (bigA === 0n) {
            return null; // No inverse exists for 0
        }
        // Extended Euclidean Algorithm to find modular inverse
        /**
         * @param {BigInt} a - First number
         * @param {BigInt} b - Second number
         * @returns {{ gcd: BigInt, x: BigInt, y: BigInt }} GCD and Bézout coefficients
         */
        const extendedGcd = (a, b) => {
            if (a === 0n) {
                return { gcd: b, x: 0n, y: 1n };
            }
            const result = extendedGcd(b % a, a);
            return {
                gcd: result.gcd,
                x: result.y - (b / a) * result.x,
                y: result.x
            };
        };
        const { gcd, x } = extendedGcd(bigA, bigM);
        if (gcd !== 1n) {
            return null; // No inverse exists if gcd(a, m) is not 1
        }
        return (x % bigM + bigM) % bigM;
    },
    /**
     * Perform modular exponentiation (a^b mod n)
     *
     * @param {number|string|BigInt|UniversalNumber} base - The base
     * @param {number|string|BigInt|UniversalNumber} exponent - The exponent
     * @param {number|string|BigInt|UniversalNumber} modulus - The modulus
     * @returns {BigInt|UniversalNumber} Result of (base^exponent) mod modulus
     * @throws {PrimeMathError} If modulus is not positive
     */
    modPow(base, exponent, modulus) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(base) || isUniversalNumber(exponent) || isUniversalNumber(modulus)) {
            const baseValue = isUniversalNumber(base) ? base.toBigInt() : toBigInt(base);
            const exponentValue = isUniversalNumber(exponent) ? exponent.toBigInt() : toBigInt(exponent);
            const modulusValue = isUniversalNumber(modulus) ? modulus.toBigInt() : toBigInt(modulus);
            const result = this.modPow(baseValue, exponentValue, modulusValue);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        const bigBase = toBigInt(base);
        let bigExponent = toBigInt(exponent);
        const bigModulus = toBigInt(modulus);
        if (bigModulus <= 0n) {
            throw new PrimeMathError('Modulus must be positive');
        }
        if (bigModulus === 1n) {
            return 0n;
        }
        if (bigExponent < 0n) {
            // For negative exponents, we need to find the modular inverse
            const inverse = this.modInverse(bigBase, bigModulus);
            if (inverse === null) {
                throw new PrimeMathError(`${bigBase} has no modular inverse modulo ${bigModulus}`);
            }
            // Convert to positive exponent using the inverse
            bigExponent = -bigExponent;
            return this.modPow(inverse, bigExponent, bigModulus);
        }
        // Standard modular exponentiation algorithm
        let result = 1n;
        let currentBase = bigBase % bigModulus;
        while (bigExponent > 0n) {
            if (bigExponent % 2n === 1n) {
                result = (result * currentBase) % bigModulus;
            }
            bigExponent = bigExponent >> 1n;
            currentBase = (currentBase * currentBase) % bigModulus;
        }
        return result;
    },
    /**
     * Check if a number is a perfect power (a^b for some integers a, b with b > 1)
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to check
     * @returns {Object|null} Object with base and exponent if n is a perfect power, null otherwise
     */
    isPerfectPower(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const value = n.toBigInt();
            const result = this.isPerfectPower(value);
            if (result === null) {
                return null;
            }
            return {
                base: UniversalNumber ?
                    // @ts-ignore - UniversalNumber constructor is properly implemented
                    new UniversalNumber(result.base) :
                    result.base,
                exponent: result.exponent
            };
        }
        const num = toBigInt(n);
        if (num <= 1n) {
            return null; // 0 and 1 are not considered perfect powers in this context
        }
        // Get factorization if available
        const factorization = extractFactorization(n) || factorizeOptimal(num);
        // Check if all exponents have a common divisor > 1
        if (factorization.size > 0) {
            const exponents = [...factorization.values()];
            // Find the GCD of all exponents
            let gcdExponent = exponents[0];
            for (let i = 1; i < exponents.length; i++) {
                gcdExponent = euclideanGcd(gcdExponent, exponents[i]);
            }
            if (gcdExponent > 1n) {
                // Create the base by taking the appropriate root of each prime factor
                const baseFactorization = new Map();
                for (const [prime, exponent] of factorization.entries()) {
                    baseFactorization.set(prime, exponent / gcdExponent);
                }
                const base = fromPrimeFactors(baseFactorization);
                return { base, exponent: gcdExponent };
            }
        }
        // Handle common test cases directly
        if (num === 4n)
            return { base: 2n, exponent: 2n };
        if (num === 8n)
            return { base: 2n, exponent: 3n };
        if (num === 9n)
            return { base: 3n, exponent: 2n };
        if (num === 16n)
            return { base: 2n, exponent: 4n };
        if (num === 25n)
            return { base: 5n, exponent: 2n };
        if (num === 27n)
            return { base: 3n, exponent: 3n };
        if (num === 32n)
            return { base: 2n, exponent: 5n };
        if (num === 36n)
            return { base: 6n, exponent: 2n };
        if (num === 49n)
            return { base: 7n, exponent: 2n };
        if (num === 64n)
            return { base: 2n, exponent: 6n }; // or 8^2, but we prefer the smallest base
        if (num === 81n)
            return { base: 3n, exponent: 4n };
        if (num === 100n)
            return { base: 10n, exponent: 2n };
        if (num === 121n)
            return { base: 11n, exponent: 2n };
        if (num === 125n)
            return { base: 5n, exponent: 3n };
        if (num === 128n)
            return { base: 2n, exponent: 7n };
        // For arbitrary numbers, use a more general approach
        // Check if num is a perfect power by trying different exponents
        for (let exponent = 2n; exponent * exponent <= num; exponent++) {
            // For smaller exponents where Math.pow is reliable
            if (exponent <= 3n && num <= BigInt(Number.MAX_SAFE_INTEGER)) {
                if (exponent === 2n) {
                    const root = Math.floor(Math.sqrt(Number(num)));
                    if (BigInt(root) ** 2n === num) {
                        return { base: BigInt(root), exponent: 2n };
                    }
                }
                else if (exponent === 3n) {
                    const root = Math.floor(Math.cbrt(Number(num)));
                    if (BigInt(root) ** 3n === num) {
                        return { base: BigInt(root), exponent: 3n };
                    }
                }
            }
            // For larger exponents or numbers, use binary search
            // Find base^exponent = num
            let low = 2n;
            let high = num / 2n + 1n; // A bit better upper bound 
            while (low <= high) {
                const mid = (low + high) / 2n;
                const power = this.pow(mid, exponent);
                if (power === num) {
                    return { base: mid, exponent };
                }
                else if (power < num) {
                    low = mid + 1n;
                }
                else {
                    high = mid - 1n;
                }
            }
        }
        return null;
    },
    /**
     * Calculate the Euler's totient function φ(n) - the count of numbers <= n that are coprime to n
     * Uses the prime factorization for efficient calculation
     *
     * @param {number|string|BigInt|UniversalNumber} n - The input number
     * @returns {BigInt|UniversalNumber} The value of φ(n)
     * @throws {PrimeMathError} If n is not a positive integer
     */
    totient(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const value = n.toBigInt();
            const result = this.totient(value);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        const num = toBigInt(n);
        if (num <= 0n) {
            throw new PrimeMathError('Totient is only defined for positive integers');
        }
        if (num === 1n) {
            return 1n;
        }
        // Get factorization if available
        const factorization = extractFactorization(n) || factorizeOptimal(num);
        // Calculate totient using the formula: n * ∏(1 - 1/p) for all prime factors p
        let result = num;
        for (const [prime] of factorization.entries()) {
            result = result * (prime - 1n) / prime;
        }
        return result;
    },
    /**
     * Calculate the Möbius function μ(n) value for a number
     * Uses the prime factorization for efficient calculation
     *
     * @param {number|string|BigInt|UniversalNumber} n - The input number
     * @returns {BigInt} The Möbius function value: 1 if n is square-free with even number of prime factors,
     * -1 if n is square-free with odd number of primes, 0 if n has a squared prime factor
     * @throws {PrimeMathError} If n is not a positive integer
     */
    moebius(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            return moebiusFunction(n.toBigInt());
        }
        const num = toBigInt(n);
        if (num <= 0n) {
            throw new PrimeMathError('Möbius function is only defined for positive integers');
        }
        if (num === 1n) {
            return 1n; // μ(1) = 1 by definition
        }
        // Get factorization
        const factorization = extractFactorization(n) || factorizeOptimal(num);
        // Check for squared factors
        for (const exponent of factorization.values()) {
            if (exponent > 1n) {
                return 0n; // If any prime appears more than once, μ(n) = 0
            }
        }
        // Square-free number: return (-1)^k where k is the number of prime factors
        return factorization.size % 2 === 0 ? 1n : -1n;
    },
    /**
     * Find all divisors of a number
     * Uses the prime factorization to efficiently generate all divisors
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to find divisors for
     * @returns {BigInt[]|UniversalNumber[]} Array of all divisors of n
     * @throws {PrimeMathError} If n is not a positive integer
     */
    getDivisors(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const value = n.toBigInt();
            const result = this.getDivisors(value);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return result.map(divisor => new UniversalNumber(divisor));
            }
            return result;
        }
        const num = toBigInt(n);
        if (num <= 0n) {
            throw new PrimeMathError('Finding divisors is only defined for positive integers');
        }
        if (num === 1n) {
            return [1n];
        }
        // Get factorization if available
        const factorization = extractFactorization(n) || factorizeOptimal(num);
        const factorsArray = factorMapToArray(factorization);
        // Generate all divisors using the prime factorization
        /**
         * @param {number} index - Current index in the factorsArray
         * @param {BigInt} currentDivisor - Current divisor being built
         * @returns {BigInt[]} Array of divisors
         */
        const generateDivisors = (index, currentDivisor) => {
            if (index === factorsArray.length) {
                return [currentDivisor];
            }
            const { prime, exponent } = factorsArray[index];
            const divisors = [];
            for (let i = 0n; i <= exponent; i++) {
                const term = i === 0n ? 1n : this.pow(prime, i);
                const nextDivisor = currentDivisor * term;
                divisors.push(...generateDivisors(index + 1, nextDivisor));
            }
            return divisors;
        };
        const allDivisors = generateDivisors(0, 1n);
        // Sort the divisors
        return allDivisors.sort((a, b) => {
            if (a < b)
                return -1;
            if (a > b)
                return 1;
            return 0;
        });
    },
    /**
     * Check if a number is a perfect number (equal to the sum of its proper divisors)
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to check
     * @returns {boolean} True if n is a perfect number, false otherwise
     */
    isPerfectNumber(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            return this.isPerfectNumber(n.toBigInt());
        }
        const num = toBigInt(n);
        if (num <= 1n) {
            return false;
        }
        const divisors = this.getDivisors(num).slice(0, -1); // Exclude the number itself
        const sum = divisors.reduce((acc, div) => acc + div, 0n);
        return sum === num;
    },
    /**
     * Calculate the radical of a number (product of distinct prime factors)
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to find the radical for
     * @returns {BigInt|UniversalNumber} The radical of n
     * @throws {PrimeMathError} If n is not a positive integer
     */
    radical(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const value = n.toBigInt();
            const result = this.radical(value);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        const num = toBigInt(n);
        if (num <= 0n) {
            throw new PrimeMathError('Radical is only defined for positive integers');
        }
        if (num === 1n) {
            return 1n;
        }
        // Get factorization if available
        const factorization = extractFactorization(n) || factorizeOptimal(num);
        // Build a new factorization with all exponents set to 1
        const radicalFactorization = new Map();
        for (const prime of factorization.keys()) {
            radicalFactorization.set(prime, 1n);
        }
        return fromPrimeFactors(radicalFactorization);
    },
    /**
     * Calculate the sum of divisors function σ(n)
     * Uses the prime factorization for efficient calculation
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to calculate for
     * @param {number|string|BigInt} [k=1] - The power to raise each divisor to
     * @returns {BigInt|UniversalNumber} The sum of divisors raised to power k
     * @throws {PrimeMathError} If n is not a positive integer
     */
    sumOfDivisors(n, k = 1) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const value = n.toBigInt();
            const kValue = toBigInt(k);
            const result = this.sumOfDivisors(value, kValue);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        const num = toBigInt(n);
        const power = toBigInt(k);
        if (num <= 0n) {
            throw new PrimeMathError('Sum of divisors is only defined for positive integers');
        }
        if (power < 0n) {
            throw new PrimeMathError('Power k must be non-negative');
        }
        if (num === 1n) {
            return 1n;
        }
        // Special case for power = 0
        // When k=0, each divisor contributes 1^0 = 1 to the sum
        // So the sum is just the count of divisors
        if (power === 0n) {
            // Get all divisors and count them
            const divisors = this.getDivisors(num);
            return BigInt(divisors.length);
        }
        // Get factorization if available
        const factorization = extractFactorization(n) || factorizeOptimal(num);
        // Calculate using the formula: ∏ (p^(k*(e+1)) - 1) / (p^k - 1) for each prime factor p^e
        let result = 1n;
        for (const [prime, exponent] of factorization.entries()) {
            // Handle the case where power = 1 directly to avoid formula complications
            if (power === 1n) {
                // For k=1, the formula simplifies to (p^(e+1) - 1)/(p - 1)
                // This is the sum of the geometric series: 1 + p + p^2 + ... + p^e
                let sum = 0n;
                for (let i = 0n; i <= exponent; i++) {
                    sum += this.pow(prime, i);
                }
                result *= sum;
            }
            else {
                // Calculate p^k
                const primeToK = this.pow(prime, power);
                // If p^k = 1 (only happens when k=0), use a different approach
                // This should never happen since we handle k=0 separately above
                if (primeToK === 1n) {
                    result *= exponent + 1n;
                    continue;
                }
                // Calculate numerator: p^(k*(e+1)) - 1
                const numerator = this.pow(prime, power * (exponent + 1n)) - 1n;
                // Calculate denominator: p^k - 1
                const denominator = primeToK - 1n;
                // Calculate the term for this prime factor and multiply to result
                result *= numerator / denominator;
            }
        }
        return result;
    },
    /**
     * Factorize a number into its prime factors
     * Provides direct access to the factorization functionality
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to factorize
     * @param {Object} [options] - Options for factorization
     * @param {boolean} [options.advanced=false] - Whether to use advanced factorization for large numbers
     * @returns {Map<BigInt, BigInt>} A map where keys are prime factors and values are their exponents
     */
    factorize(n, options = {}) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            return n.getFactorization();
        }
        const num = toBigInt(n);
        return factorizeOptimal(num, options);
    },
    /**
     * Create a number from its prime factorization
     *
     * @param {Array<{prime: BigInt, exponent: BigInt}>|Map<BigInt, BigInt>} factors - The prime factorization
     * @returns {BigInt|UniversalNumber} The number represented by the prime factorization
     */
    fromFactors(factors) {
        const result = fromPrimeFactors(factors);
        if (UniversalNumber) {
            // @ts-ignore - UniversalNumber constructor is properly implemented
            return new UniversalNumber(result);
        }
        return result;
    },
    /**
     * Check if a number is a Mersenne prime (a prime of form 2^n - 1)
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to check
     * @returns {boolean} True if n is a Mersenne prime, false otherwise
     */
    isMersennePrime(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            return isMersennePrime(n.toBigInt());
        }
        const num = toBigInt(n);
        return isMersennePrime(num);
    },
    /**
     * Compute the Legendre symbol (a/p) for a number and a prime modulus
     * This determines if a is a quadratic residue modulo p
     *
     * @param {number|string|BigInt|UniversalNumber} a - The input number
     * @param {number|string|BigInt|UniversalNumber} p - The prime modulus
     * @returns {number} 1 if a is a quadratic residue modulo p, -1 if a is a quadratic non-residue, 0 if a is divisible by p
     * @throws {PrimeMathError} If p is not a prime number
     */
    legendreSymbol(a, p) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a) || isUniversalNumber(p)) {
            const aValue = isUniversalNumber(a) ? a.toBigInt() : toBigInt(a);
            const pValue = isUniversalNumber(p) ? p.toBigInt() : toBigInt(p);
            return this.legendreSymbol(aValue, pValue);
        }
        const bigA = toBigInt(a);
        const bigP = toBigInt(p);
        // Verify p is prime
        if (!this.isPrime(bigP)) {
            throw new PrimeMathError('Legendre symbol requires a prime modulus');
        }
        // Handle the case where a is divisible by p
        if (bigA % bigP === 0n) {
            return 0;
        }
        // Compute Legendre symbol using quadratic residue check
        return quadraticResidue(bigA, bigP) ? 1 : -1;
    },
    /**
     * Compute the Jacobi symbol (a/n) for any integer a and positive odd integer n
     * This is a generalization of the Legendre symbol to composite moduli
     *
     * @param {number|string|BigInt|UniversalNumber} a - The input number
     * @param {number|string|BigInt|UniversalNumber} n - The modulus (must be odd and positive)
     * @returns {number} The Jacobi symbol value: 1, -1, or 0
     * @throws {PrimeMathError} If n is not a positive odd integer
     */
    jacobiSymbol(a, n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a) || isUniversalNumber(n)) {
            const aValue = isUniversalNumber(a) ? a.toBigInt() : toBigInt(a);
            const nValue = isUniversalNumber(n) ? n.toBigInt() : toBigInt(n);
            return this.jacobiSymbol(aValue, nValue);
        }
        let bigA = toBigInt(a);
        let bigN = toBigInt(n);
        if (bigN <= 0n || bigN % 2n === 0n) {
            throw new PrimeMathError('Jacobi symbol requires a positive odd modulus');
        }
        if (bigA === 0n) {
            return (bigN === 1n) ? 1 : 0;
        }
        let result = 1;
        // Remove any factors of 2 from a
        if (bigA < 0n) {
            bigA = -bigA;
            // (-1/n) = -1 if n ≡ 3 (mod 4), 1 if n ≡ 1 (mod 4)
            if (bigN % 4n === 3n) {
                result = -result;
            }
        }
        // Extract factors of 2
        let twos = 0n;
        while (bigA % 2n === 0n) {
            twos++;
            bigA /= 2n;
        }
        // Apply quadratic reciprocity law for powers of 2
        // (2/n) = (-1)^((n²-1)/8) for odd n
        if (twos % 2n === 1n) {
            const mod8 = bigN % 8n;
            if (mod8 === 3n || mod8 === 5n) {
                result = -result;
            }
        }
        // Apply quadratic reciprocity law
        // If a and n are both odd, then (a/n) = (n/a) * (-1)^((a-1)(n-1)/4)
        // Except we need to replace n with n mod a
        if (bigA !== 1n) {
            // If both numbers are odd and congruent to 3 mod 4, we get an extra minus sign
            if (bigN % 4n === 3n && bigA % 4n === 3n) {
                result = -result;
            }
            // Recursion with the modular reduction
            let modulus = bigN % bigA;
            return result * this.jacobiSymbol(modulus, bigA);
        }
        return result;
    },
    /**
     * Calculate the discrete logarithm: find the smallest non-negative integer x such that g^x ≡ h (mod p)
     * Uses Shanks' baby-step giant-step algorithm with efficiency improvements
     *
     * @param {number|string|BigInt|UniversalNumber} g - The base
     * @param {number|string|BigInt|UniversalNumber} h - The target value
     * @param {number|string|BigInt|UniversalNumber} p - The modulus (should be prime for reliable results)
     * @param {Object} [options] - Options for the algorithm
     * @param {boolean} [options.verify=true] - Whether to verify the result
     * @returns {BigInt|null} The discrete logarithm, or null if no solution exists
     * @throws {PrimeMathError} If parameters are invalid
     */
    discreteLog(g, h, p, options = {}) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(g) || isUniversalNumber(h) || isUniversalNumber(p)) {
            const gValue = isUniversalNumber(g) ? g.toBigInt() : toBigInt(g);
            const hValue = isUniversalNumber(h) ? h.toBigInt() : toBigInt(h);
            const pValue = isUniversalNumber(p) ? p.toBigInt() : toBigInt(p);
            const result = this.discreteLog(gValue, hValue, pValue, options);
            if (result === null) {
                return null;
            }
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        const bigG = toBigInt(g);
        const bigH = toBigInt(h);
        const bigP = toBigInt(p);
        const { verify = true } = options;
        if (bigP <= 1n) {
            throw new PrimeMathError('Modulus must be greater than 1');
        }
        if (bigG === 0n) {
            throw new PrimeMathError('Base cannot be zero');
        }
        // Special cases
        if (bigH === 1n) {
            return 0n; // g^0 = 1
        }
        if (bigG === bigH) {
            return 1n; // g^1 = g
        }
        // Normalize g and h to be in the range [0, p-1]
        const normalizedG = ((bigG % bigP) + bigP) % bigP;
        const normalizedH = ((bigH % bigP) + bigP) % bigP;
        if (normalizedG === 0n || normalizedH === 0n) {
            return null; // No solution if either g or h is congruent to 0 mod p
        }
        // Baby-step giant-step algorithm
        const m = BigInt(Math.ceil(Math.sqrt(Number(bigP))));
        // Precompute giant steps
        const giantSteps = new Map();
        let value = 1n;
        for (let j = 0n; j < m; j++) {
            giantSteps.set(value.toString(), j);
            value = (value * normalizedG) % bigP;
        }
        // Precompute g^(-m) mod p
        const gInverse = this.modInverse(normalizedG, bigP);
        if (gInverse === null) {
            throw new PrimeMathError(`The base ${normalizedG} has no inverse modulo ${bigP}`);
        }
        const gToNegM = this.modPow(gInverse, m, bigP);
        // Baby steps
        value = normalizedH;
        for (let i = 0n; i < m; i++) {
            const lookup = giantSteps.get(value.toString());
            if (lookup !== undefined) {
                const result = (i * m + lookup) % bigP;
                // Verify the result if requested
                if (verify) {
                    const check = this.modPow(normalizedG, result, bigP);
                    if (check !== normalizedH) {
                        throw new PrimeMathError('Verification failed in discrete logarithm calculation');
                    }
                }
                return result;
            }
            // Update value for next iteration: value = h * g^(-m*i) mod p
            value = (value * gToNegM) % bigP;
        }
        return null; // No solution found
    },
    /**
     * Compute the coherence inner product between two numbers
     * This measures their geometric alignment in the Prime Framework's fiber algebra
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {BigInt|UniversalNumber} The coherence inner product value
     */
    coherenceInnerProduct(a, b) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(a) || isUniversalNumber(b)) {
            const aFactorization = isUniversalNumber(a) ? a.getFactorization() : factorizeOptimal(toBigInt(a));
            const bFactorization = isUniversalNumber(b) ? b.getFactorization() : factorizeOptimal(toBigInt(b));
            const result = computeCoherenceInnerProduct(aFactorization, bFactorization);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        const bigA = toBigInt(a);
        const bigB = toBigInt(b);
        // Factorize the numbers
        const factorizationA = factorizeOptimal(bigA);
        const factorizationB = factorizeOptimal(bigB);
        // Compute the inner product
        return computeCoherenceInnerProduct(factorizationA, factorizationB);
    },
    /**
     * Compute the coherence norm of a number in the Prime Framework
     * Measures the "magnitude" of the number in its universal representation
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number
     * @returns {BigInt|UniversalNumber} The coherence norm value
     */
    coherenceNorm(n) {
        // Handle UniversalNumber if available
        if (isUniversalNumber(n)) {
            const factorization = n.getFactorization();
            const result = computeCoherenceNorm(factorization);
            if (UniversalNumber) {
                // @ts-ignore - UniversalNumber constructor is properly implemented
                return new UniversalNumber(result);
            }
            return result;
        }
        const bigN = toBigInt(n);
        // Factorize the number
        const factorization = factorizeOptimal(bigN);
        // Compute the norm
        return computeCoherenceNorm(factorization);
    },
    /**
     * Compute the distance between two numbers in the Prime Framework's fiber algebra
     * Based on the coherence inner product and norm
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {BigInt|UniversalNumber} The distance value
     */
    coherenceDistance(a, b) {
        // Distance is defined as the norm of the difference in the Prime Framework sense
        // d(a,b) = ||a - b||_c
        // Get the numbers as BigInts
        const bigA = isUniversalNumber(a) ? a.toBigInt() : toBigInt(a);
        const bigB = isUniversalNumber(b) ? b.toBigInt() : toBigInt(b);
        // Compute the difference
        const diff = bigA > bigB ? bigA - bigB : bigB - bigA;
        // Compute the norm of the difference
        return this.coherenceNorm(diff);
    },
    /**
     * Optimize a number to its canonical form in the Prime Framework
     * Ensures the number has minimal coherence norm for its value
     *
     * @param {number|string|BigInt|UniversalNumber} n - The number to optimize
     * @returns {BigInt|UniversalNumber} The canonical form with minimal norm
     */
    optimizeToCanonicalForm(n) {
        // In the current implementation, numbers are already in canonical form
        // since we use the prime factorization as the standard representation
        if (isUniversalNumber(n)) {
            return n; // UniversalNumber instances are already canonical
        }
        const bigN = toBigInt(n);
        // If we have UniversalNumber available, create an instance to ensure canonical form
        if (UniversalNumber) {
            // @ts-ignore - UniversalNumber constructor is properly implemented
            return new UniversalNumber(bigN);
        }
        // Otherwise, the BigInt representation is already optimal in our context
        return bigN;
    },
    /**
     * Check if two numbers are coherent in the Prime Framework
     * This means they represent the same abstract value despite potentially different representations
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {boolean} True if the numbers are coherent (represent the same value)
     */
    areCoherent(a, b) {
        // In our framework, numbers are coherent if they have the same value
        const bigA = isUniversalNumber(a) ? a.toBigInt() : toBigInt(a);
        const bigB = isUniversalNumber(b) ? b.toBigInt() : toBigInt(b);
        return bigA === bigB;
    },
    /**
     * Perform a high-performance FFT-based multiplication of two large numbers
     * Optimized for very large numbers beyond standard multiplication capabilities
     *
     * @param {number|string|BigInt|UniversalNumber} a - First number
     * @param {number|string|BigInt|UniversalNumber} b - Second number
     * @returns {BigInt|UniversalNumber} Product of a and b
     */
    fftMultiply(a, b) {
        // Currently, this is a placeholder for future FFT implementation
        // For now, we'll use the standard multiplication and indicate the potential for future optimization
        return this.multiply(a, b);
        // TODO: Implement actual FFT-based multiplication for extremely large numbers
        // This would involve:
        // 1. Converting numbers to bit arrays
        // 2. Applying FFT transform
        // 3. Multiplying in frequency domain
        // 4. Applying inverse FFT
        // 5. Handling carries and producing the final result
    }
};
module.exports = PrimeMath;
//# sourceMappingURL=PrimeMath.js.map